/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

/*
 * ProofPanel.java
 *
 * Created on 16 oct. 2010, 00:30:55
 */

package panda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import javax.swing.JViewport;
import javax.swing.Timer;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


/**
 * This is a panel where we write proofs
 * @author Ancmin
 */
public class ProofPanel extends ZoomablePanel {

    /**
     * in order to be able to cancel/redo
     */
    private final CommandsManager commandsManager = new CommandsManager(this);

    //list of nodes
    private final ArrayList<ProofFormulaNode> nodes = new ArrayList<ProofFormulaNode>();

    //current node under the mouse pointer
    private ProofFormulaNode currentNode = null;

    //set of nodes that are selected
    private final Set<ProofFormulaNode> nodesSelected = new HashSet<ProofFormulaNode>();

    //node that is currently moved
    private ProofFormulaNode currentNodeMoved = null;
    private ActionListener actionListenerSelectionChange = null;
    private ActionListener actionListenerModification = null;

    //points for drawing a line (Bottom-up deduction, for instance, from A and B I deduce A and B)
    private Point dessinTraitPoint1 = null;
    private Point dessinTraitPoint2 = null;

    //timer for animation (node explosion, nodes which move)
    private final Timer timerAnimation = new Timer(50, new ActionListener() {

        public void actionPerformed(ActionEvent e) {
            for(ProofFormulaNode node : nodes)
            {
                node.animate();
            }
            repaint();
        }
    });

    //list of rectangle of surprise, that is to say rectangles where an explosion has occured (due to node
    //deletion for instance)
    private final ArrayList<AnimatedRectangle> surprisedRectangles = new ArrayList<AnimatedRectangle>();


    private Point mouseDragLastPoint;
    private Point currentNodePointMilieuHautBeginMouseDrag;
    
    public void setActionListenerSelectionChange(ActionListener actionListenerSelectionChange) {
        this.actionListenerSelectionChange = actionListenerSelectionChange;
    }

    public void setActionListenerModification(ActionListener actionListenerModification) {
        this.actionListenerModification = actionListenerModification;
    }


    
/**
 * execute the command. (you create object of type Command that contains all information to perform an action)
 * and then you apply this action with this method
 * Example: commandExecute(new CommandAddNode(n)) in order to add a node
 * @param command
 */
    public void commandExecute(Command command)
    {
        commandsManager.execute(command);
    }


/**
 * undo, cancel ONE command
 */
    public void commandUnDo()
    {
        commandsManager.undoLastCommand();
        calculer();
        repaint();


    }


    /**
     * redo the last command
     */
    public void commandReDo()
    {
        commandsManager.redoNextCommand();
        calculer();
        repaint();
    }

    

    /** Creates new form ProofPanel */
    public ProofPanel() {
        initComponents();
      //  zoomMouseListenerAdapt();
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        popupMenuNode = new javax.swing.JPopupMenu();
        mnuCacherAfficher = new javax.swing.JMenuItem();

        popupMenuNode.setName("popupMenuNode"); // NOI18N

        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(panda.MrnatureApp.class).getContext().getResourceMap(ProofPanel.class);
        mnuCacherAfficher.setText(resourceMap.getString("mnuCacherAfficher.text")); // NOI18N
        mnuCacherAfficher.setName("mnuCacherAfficher"); // NOI18N
        mnuCacherAfficher.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                mnuCacherAfficherActionPerformed(evt);
            }
        });
        popupMenuNode.add(mnuCacherAfficher);

        setBackground(resourceMap.getColor("Form.background")); // NOI18N
        addMouseWheelListener(new java.awt.event.MouseWheelListener() {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) {
                formMouseWheelMoved(evt);
            }
        });
        addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(java.awt.event.MouseEvent evt) {
                formMouseClicked(evt);
            }
            public void mousePressed(java.awt.event.MouseEvent evt) {
                formMousePressed(evt);
            }
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                formMouseReleased(evt);
            }
        });
        addMouseMotionListener(new java.awt.event.MouseMotionAdapter() {
            public void mouseDragged(java.awt.event.MouseEvent evt) {
                formMouseDragged(evt);
            }
            public void mouseMoved(java.awt.event.MouseEvent evt) {
                formMouseMoved(evt);
            }
        });
        addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyPressed(java.awt.event.KeyEvent evt) {
                formKeyPressed(evt);
            }
        });

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 528, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 370, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents

    private void formMouseClicked(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_formMouseClicked
        if(evt.getClickCount() > 1)
        {
            FormulaDialog f;
            if(currentNode != null)
                f = new FormulaDialog(null, true, currentNode.getFormula(), getFormulasHypotheses());
                 
            else
                 f = new FormulaDialog(null, true, getFormulasHypotheses());
            
            f.setVisible(true);
            if(f.isOK())
            {
                if(currentNode != null)
                    commandExecute(new CommandNodeChangeFormula(currentNode, f.getFormula()));
                else
                    commandExecute(new CommandAddNode(new ProofFormulaNodeNatDet(evt.getPoint(), f.getFormula())));
                repaint();
            }
            
        }
        else if(evt.getButton() == MouseEvent.BUTTON3)
        {
            popupMenuNode.show(this, (int) getXPhysical(evt.getX()), (int) getYPhysical(evt.getY()));
        }
    }





/**
 *
 * @return l'ensemble des formules hypothèses (à déchargeables quelque part ou apparaissant
 * dans une feuille quelque part dans les arbres de preuve)
 * Utilité : pour la boîte de dialogue FormulaDialog
 */
    public Set<Formula> getFormulasHypotheses()
    {
        Set<Formula> F = new HashSet<Formula>();
        for(ProofFormulaNode node : nodes)
        {
            F.addAll(((ProofFormulaNodeNatDet) node).getHypothesesGeneres());
        }
        return F;
    }


    /**
     * fait en sorte que l'affichage est joli (antialiasing etc.)
     * @param g2
     */
    private void graphics2Dconfigure(Graphics2D g2)
    {
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
        RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setRenderingHint(RenderingHints.KEY_RENDERING,
        RenderingHints.VALUE_RENDER_QUALITY);
        g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
                RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED);
    }



    /**
 * entame un affiche transparent
 * @param g2
 */
    protected void drawTransparent(Graphics2D g2)
    {
        g2.setComposite(java.awt.AlphaComposite.getInstance(
                java.awt.AlphaComposite.SRC_OVER,Math.min(
                    0.2f, //0 transparent à fond,  1 = opaque
                    1.0f ) )) ;
    }



    /**
     * s'occupe de l'affichage
     * @param g
     */
    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        graphics2Dconfigure((Graphics2D) g);

        surprisedRectanglesClean();
        animationTimerVerifier();



        for(ProofFormulaNode nodeSelected : nodesSelected)
            nodeSelected.drawAsSelected(g);


        /**
         * draw the line for deduce something by mouse by bottom-up
         */
        if(isTraitSouris())
        {
            ((Graphics2D) g).setStroke(new BasicStroke(2.0f));

            g.drawLine(getTraitSourisXLeft(), getTraitSourisY(),
                       getTraitSourisXRight(), getTraitSourisY());

            for(ProofFormulaNode node : getChildrenTraitSouris())
            {
                node.drawAsSelected(g);
            }
        }


        /**
         * draw the nodes
         */
        for(ProofFormulaNode node : getShownNodes())
        {
            if(node.noFather())
                 node.drawNodeAndTree(g);
        }




        if(currentNode != null)
             currentNode.drawAsCurrent(g);

        

        if(currentNodeMoved != null)
        for(ProofFormulaNode node : getShownNodes())
        {
            if(node.askToBeFatherOf(currentNodeMoved))
                node.drawTraitGros(g);

            if(currentNodeMoved.askToBeFatherOf(node) & node.getFather() == null)
                currentNodeMoved.drawTraitGros(g);
        }


        
       

        
        /*
         * show the surprised rectangle (animation)
         */
        for(int i = 0; i < surprisedRectangles.size(); i++)
        {
            surprisedRectangles.get(i).draw(g);
            
        }


        /*
         * show the rectangle for selecting nodes
         */
        if(isSelectionRectangle())
        {
            Rectangle r = getSelectionRectangle();
            drawTransparent((Graphics2D) g);
            g.setColor(new Color(0.5f, 0.5f, 1.0f));
            g.fillRect(r.x, r.y, r.width, r.height);
            g.drawRect(r.x, r.y, r.width, r.height);

        }
        
    }//GEN-LAST:event_formMouseClicked

/**
 *
 * @param p
 * @return the node that is under the point p.
 * if no node, return null.
 */
    private ProofFormulaNode getNode(Point p)
    {
        for(ProofFormulaNode node : getShownNodes())
        {
            if(node.isOverFormula(p))
                  return node;
        }
        return null;
    }


    private void formMouseMoved(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_formMouseMoved
        currentNode = getNode(evt.getPoint());
        currentNodeMoved = null;
        repaint();
    }//GEN-LAST:event_formMouseMoved


    
    private void formMouseDragged(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_formMouseDragged
        if(currentNode != null)
        {
            for(ProofFormulaNode node : getSelectedNodes())
            if(noAncestorSelected(node))
            {
                node.translate(evt.getPoint().x - mouseDragLastPoint.x,
                               evt.getPoint().y - mouseDragLastPoint.y);
            }

            mouseDragLastPoint =  evt.getPoint();
            
          //  currentNode.setPointMilieuHaut(p);
            currentNodeMoved = currentNode;

            layoutDo();
            
            repaint();

        }
        else
        {
            dessinTraitPoint2 = evt.getPoint();
            repaint();
        }
    }//GEN-LAST:event_formMouseDragged


    
    private void formMousePressed(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_formMousePressed
        if(currentNode != null)
        {
            mouseDragLastPoint = evt.getPoint();
            currentNodePointMilieuHautBeginMouseDrag = (Point) currentNode.getPointMilieuHaut().clone();

            if(!evt.isControlDown() & !isNodeSelected(currentNode))
                noSelection();

            addNodeSelected(currentNode);
        }
        else
        {
            setNodeSelected(null);
            dessinTraitPoint1 = evt.getPoint();
            dessinTraitPoint2 = evt.getPoint();
        }
        repaint();
    }//GEN-LAST:event_formMousePressed







    private void calculer()
    {
        positionsCompute();
        actionListenerModification.actionPerformed(null);
        selectionChange();
    }


    private void formMouseReleased(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_formMouseReleased

        if(currentNode != null)
        {
            Point lastPositioncurrentNode = currentNodePointMilieuHautBeginMouseDrag;
            
            for(ProofFormulaNode node : getShownNodes())
            {
                if(node.askToBeFatherOf(currentNode))
                {
                    CommandComposees c = new CommandComposees();

                    
                    if(node.noChildren() & node.getFormula().equals(currentNode.getFormula()))
                        //fusionne node et currentNode
                    {
                        ProofFormulaNode fatherOfNode = node.getFather();

                        if(fatherOfNode == null)
                        {
                            addSurprisedRectangle(node.getRectangleTrait());
                            c.commandAdd(new CommandRemoveNode(node));
                        }
                        else
                        {
                            c.commandAdd(new CommandRemoveNode(node));
                            c.commandAdd(new CommandNodeAddChild(fatherOfNode, currentNode, lastPositioncurrentNode));
                           
                            addSurprisedRectangle(fatherOfNode.getRectangleTrait());
                        }



                    }
                    else
                    {
                        if(currentNode.getFather() != node)
                        {
                            c.commandAdd(new CommandNodeAddChild(node, currentNode, lastPositioncurrentNode));
                            addSurprisedRectangle(node.getRectangleTrait());
                        }
                         

                    }

                    commandExecute(c);
                    calculerHypothesesTemporaires();
                    calculer();

                    repaint();
                    return;
                }
                
            }



            CommandComposees c = new CommandComposees();
            c.commandAdd(new CommandNodeDeconnectFromFather(currentNode, lastPositioncurrentNode, currentNode.getPointMilieuHaut()));
            


            for(ProofFormulaNode node : getShownNodes())
            {
                if(currentNode.askToBeFatherOf(node) & node.getFather() == null)
                {

                    c.commandAdd(new CommandNodeAddChild(currentNode, node, lastPositioncurrentNode));
                    commandExecute(c);

                    calculerHypothesesTemporaires();
                    calculer();

                    repaint();
                    return;
                }

            }


            commandExecute(c);
        }
        else if(isTraitSouris())
        {
            applyRuleByTraitSouris();
            TraitSourisAucun();
            
        }
        else if(isSelectionRectangle())
        {
            selectionnerRectangle(getSelectionRectangle());
            TraitSourisAucun();
        }

        
        selectionChange();
        actionListenerModification.actionPerformed(null);
        calculer();
        repaint();

    }//GEN-LAST:event_formMouseReleased

    private void formKeyPressed(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_formKeyPressed

    }//GEN-LAST:event_formKeyPressed

    private void mnuCacherAfficherActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_mnuCacherAfficherActionPerformed
        for(ProofFormulaNode node : getSelectedNodes())
        {
            if(node.isTreeHidden())
                node.treeShowAll();
            else
                nodeTreeHide(node);
        }

        calculer();
        repaint();
    }//GEN-LAST:event_mnuCacherAfficherActionPerformed

    private void formMouseWheelMoved(java.awt.event.MouseWheelEvent evt) {//GEN-FIRST:event_formMouseWheelMoved
        
    }//GEN-LAST:event_formMouseWheelMoved



    public void nodeAdd(final ProofFormulaNode node)
    {
        nodes.add(node);
        calculerHypothesesTemporaires();
        actionListenerModification.actionPerformed(null);
        addSurprisedRectangle(node.getRectangleFormula());
        
    }



    public void nodeRemove(final ProofFormulaNode node)
    {
        if(node == null)
            System.out.println("nodeRemove reçoit un node null");

        node.deconnectFromAll();
        addSurprisedRectangle(node.getRectangleFormula());
        nodes.remove(node);
        nodesSelected.remove(node);
        selectionChange();
        if(node == currentNode)
            currentNode = null;

        actionListenerModification.actionPerformed(null);
    }




    public void nodeTreeHide(final ProofFormulaNode node)
    {
        node.treeHide();
        nodesSelected.removeAll(node.getChildrenAndGrandChildrenEtc());
    }


/**
 *
 * @return the selected node.
 * Returns null if:
 * -there is no selected node;
 * - there are strictly more that one selected node
 */
    public ProofFormulaNodeNatDet getSelectedNode()
    {
        if( nodesSelected.size() > 1 )
            return null;
        for(ProofFormulaNode node : nodesSelected)
        {
            return (ProofFormulaNodeNatDet) node;
        }
        return  null;
    }


/**
 *
 * @return the set of selected nodes
 * if empty returns an empty ArrayList
 */
    public ArrayList<ProofFormulaNode> getSelectedNodes()
    {
        ArrayList<ProofFormulaNode> A = new ArrayList<ProofFormulaNode>(nodesSelected);
        Collections.sort(A, new ComparatorProofFormulaNodeX());
        return A;

    }




    /**
     * initialise tout
     */
    public void clear() {
        nodes.clear();
        noSelection();
        currentNode = null;
        commandsManager.clear();
        repaint();
    }


    /**
     * compute the number of all temporary hypothesis
     */
    private void calculerHypothesesTemporaires() {
        ProofFormulaNodeNatDet.numerotationHypotheseTemporaireInitialiser();

        if(nodes == null)
            return;

        for(ProofFormulaNode node : nodes)
            ((ProofFormulaNodeNatDet) node).hypothesesTemporairesInitialiser();
        
        for(ProofFormulaNode node : nodes)
        {
            if(node.getFather() == null)
                ((ProofFormulaNodeNatDet) node).calculerHypothesesTemporaires();
        }
    }


    
    @Override
    public void repaint() {
        super.repaint();
        calculerHypothesesTemporaires();
    }



    /**
     * the methods sees to it that the set of nodes is the singleton with
     * nodeSelected
     * @param nodeSelected
     */
    public void setNodeSelected(ProofFormulaNode nodeSelected) {
        noSelection();

        if(nodeSelected != null)
             nodesSelected.add(nodeSelected);

        selectionChange();
        
    }


    /**
     *
     * @return the preferred size for the panel considering the set of
     * nodes contained by the panel
     */
    @Override
    public Dimension getPreferredSize() {
        int xmin = 0;
        int ymin = 0;
        int xmax = 0;
        int ymax = 0;
        for(ProofFormulaNode node : getShownNodes())
        {
            if(node.getXLeft() < xmin)
                xmin = node.getXLeft();

            if(node.getXRight() > xmax)
                xmax = node.getXRight();

            if(node.getYFormulaTop() < ymin)
                ymin = node.getXLeft();

            if(node.getYBottom() > ymax)
                ymax = node.getYBottom();


        }

        return new Dimension((int) ((xmax + 100) * getZoom()),
                             (int) ((ymax + 100) * getZoom()));
    }




    private void layoutDo() {
        doLayout();
        getParent().doLayout();
    }


    public Set<HypotheseTemporaireDechargee> getHypothesesADecharger()
    {
        Set<HypotheseTemporaireDechargee> S = new HashSet<HypotheseTemporaireDechargee>();
        for(ProofFormulaNode n : nodes)
        {
            S.addAll(((ProofFormulaNodeNatDet) n).getHypothesesDechargeesIciDansCeNoeud());
        }
        return S;
    }

    private void TraitSourisAucun() {
        dessinTraitPoint1 = null;
        dessinTraitPoint2 = null;
    }


    /**
     *
     * @return true if the user is currently drawing a line of a proof
     */
    private boolean isTraitSouris()
    {
        if(dessinTraitPoint1 == null)
            return false;
        else
        {
            return Math.abs(dessinTraitPoint1.y - dessinTraitPoint2.y) < 30;
        }
    }



/**
 *
 * @return true if the user is drawing a rectangle in order to select node
 */
    private boolean isSelectionRectangle()
    {
        if(dessinTraitPoint1 == null)
            return false;
        else
            return !isTraitSouris();
    }



/**
 *
 * @return the rectangle for selecting node that the user is drawing
 * if the user is not drawing a rectangle of selection, this function may crash.
 */
    private Rectangle getSelectionRectangle()
    {
        int x1 = Math.min(dessinTraitPoint1.x, dessinTraitPoint2.x);
        int x2 = Math.max(dessinTraitPoint1.x, dessinTraitPoint2.x);

        int y1 = Math.min(dessinTraitPoint1.y, dessinTraitPoint2.y);
        int y2 = Math.max(dessinTraitPoint1.y, dessinTraitPoint2.y);
        
        return new Rectangle(x1, y1, x2 - x1, y2 - y1);
    }





    private void applyRuleByTraitSouris() {
        ArrayList<ProofFormulaNode> futureChildren = getChildrenTraitSouris();

        if(futureChildren.size() == 2)
        {
            Formula f0 = futureChildren.get(0).getFormula();
            Formula f1 = futureChildren.get(1).getFormula();

            if(f0.isNot())
                if(f1.equals(f0.getSubFormulaForNot()))
                {
                    //PandaVoice.speak("From " + f0 + " and, " + f1 + ", we have a contradiction.");
                    createRuleIntroBottom(futureChildren);
                    return;
                }

            if(f1.isNot())
                if(f0.equals(f1.getSubFormulaForNot()))
                {
                   // PandaVoice.speak("From " + f0 + ", and, " + f1 + ", we have a contradiction.");
                    createRuleIntroBottom(futureChildren);
                    return;
                }
                

            if(f0.isImply())
                if(f1.equals(f0.getSubFormulaLeft()))
                {
//                    PandaVoice.speak("As " + f1 + "is proved");
//                    PandaVoice.speak("and " + f0);
//                    PandaVoice.speak("By elimination of implication, we deduce " + f0.getSubFormulaRight());
                    

                    createCommonFatherMouseLine(futureChildren, f0.getSubFormulaRight());
                    return;
                }
                    

            if(f1.isImply())
                if(f0.equals(f1.getSubFormulaLeft()))
                {
//                    PandaVoice.speak("As " + f0 + "is proved");
//                    PandaVoice.speak("and " + f1);
//                    PandaVoice.speak("By elimination of implication, we deduce " + f1.getSubFormulaRight());
                    
                    createCommonFatherMouseLine(futureChildren, f1.getSubFormulaRight());
                    return;
                }
                    

            //PandaVoice.speak("From " + f0 + " and " + f1 + " , we simply have " + f0 + " and " + f1);
            createCommonFatherMouseLine(futureChildren, new Formula("(" + f0 + " and " + f1 + ")"));
        }
        else if(futureChildren.size() == 3)
        {
            Formula f0 = futureChildren.get(0).getFormula();
            Formula f1 = futureChildren.get(1).getFormula();
            Formula f2 = futureChildren.get(2).getFormula();

            if(f0.equals(f1))
            {
                Formula t = f2;
                f2 = f1;
                f1 = f0;
                f0 = t;
            }

            if(f0.equals(f2))
            {
                f0 = f1;
                f1 = f2;
            }

            //f0 devrait contenir le "ou"

            if(f0.isOr())
            {
                createCommonFatherMouseLine(futureChildren, f2);
            }
            
        }
        

    }




    private int getTraitSourisY()
    {
        return dessinTraitPoint1.y;
    }



    private int getTraitSourisXLeft()
    {
        return Math.min(dessinTraitPoint1.x, dessinTraitPoint2.x);
    }

    private int getTraitSourisXRight()
    {
        return Math.max(dessinTraitPoint1.x, dessinTraitPoint2.x);
    }

    private ArrayList<ProofFormulaNode> getChildrenTraitSouris()
    {
        ArrayList<ProofFormulaNode> futureChildren = new ArrayList<ProofFormulaNode>();
        for(ProofFormulaNode node : nodes)
        {
            if((node.getYBottom() < getTraitSourisY()) &
                node.getYBottom() > getTraitSourisY() - 150 &
                node.noFather() &
                node.getXMilieu() > getTraitSourisXLeft() &
                node.getXMilieu() < getTraitSourisXRight())
            {
                futureChildren.add(node);
            }
        }
        Collections.sort(futureChildren, new ComparatorProofFormulaNodeX());
        return futureChildren;
    }



/**
 *
 * @param children
 * @return the barycentre of all positions of nodes in children
 */
    private Point getBarycentre(ArrayList<ProofFormulaNode> children)
    {
        int x = 0;
        int y = 0;
        for(ProofFormulaNode child : children)
        {
            x += child.getXMilieu();
            y += child.getYBottom();
        }

        x = x / children.size();
        y = y / children.size();

        return new Point(x, y);
    }


    
    private Point getTraitSourisFatherPointMilieu()
    {
        int x = 0;
        for(ProofFormulaNode child : getChildrenTraitSouris())
        {
            x += child.getXMilieu();
        }

        x = x / getChildrenTraitSouris().size();

        return new Point(x, getTraitSourisY());
    }


    private void createRuleIntroBottom(ArrayList<ProofFormulaNode> futureChildren) {
        createCommonFatherMouseLine(futureChildren, new Formula("bottom"));
    }

/**
     * On considère des noeuds déjà présents dans la preuve (futureChildren) et on
     * crée un père commun qui contient la formule formulaFather
     * On place le père au niveau du trait tracé à la souris.
     * @param futureChildren
     * @param formulaFather
     */
    private void createCommonFatherMouseLine(ArrayList<ProofFormulaNode> futureChildren, Formula formulaFather) {
        ProofFormulaNodeNatDet father = new ProofFormulaNodeNatDet(getTraitSourisFatherPointMilieu(), formulaFather);
        commandExecute(new CommandCreateCommonFather(father, futureChildren));
        setNodeSelected(father);
    }


    /**
     * On considère des noeuds déjà présents dans la preuve (futureChildren) et on
     * crée un père commun qui contient la formule formulaFather
     * @param futureChildren
     * @param formulaFather
     */
    public void createCommonFatherBarycentre(ArrayList<ProofFormulaNode> futureChildren, Formula formulaFather) {
        ProofFormulaNodeNatDet father = new ProofFormulaNodeNatDet(getBarycentre(futureChildren), formulaFather);
        commandExecute(new CommandCreateCommonFather(father, futureChildren));
        setNodeSelected(father);
    }

    /**
     * Déselectionne tous les noeuds
     */
    private void noSelection() {
        nodesSelected.clear();
    }


    /**
     * add a node to the selection
     * @param currentNode
     */
    private void addNodeSelected(ProofFormulaNode node) {
        nodesSelected.add(node);
       
    }



    /**
     * prévient que la sélection a changée
     */
    protected void selectionChange() {
        if(actionListenerSelectionChange != null)
            actionListenerSelectionChange.actionPerformed(null);
    }



    /**
     *
     * @param nodeTest
     * @return true iff the nodec nodeTest overlaps an other node
     */
    private boolean isSuperpose(ProofFormulaNode nodeTest)
    {
        for(ProofFormulaNode node : getShownNodes())
        {
            if(node != nodeTest)
               if(node.getRectangleFormula().intersects(nodeTest.getRectangleFormula()))
                   return true;
        }
        return false;
    }

    /**
     * move the node in order to be sure it does not overlap another one
     * @param node
     */
    public void moveNodeToEnsureNoOverlaping(ProofFormulaNode node)
    {
        int xdepart = node.getXMilieu();
        int ydepart = node.getYFormulaTop();

        for(int y = ydepart; y < getHeight(); y+=20)
        {
            for(int x = xdepart; x < getWidth(); x +=20)
            {
                node.setPointMilieuHautImmediat(new Point(x, y));
                if(!isSuperpose(node))
                    return;
            }
        }
    }


    /**
     * add a rectangle for animation
     * @param rectangleTrait
     */
    private void addSurprisedRectangle(Rectangle rectangleTrait) {
        surprisedRectangles.add(new AnimatedRectangle(rectangleTrait));
        timerAnimation.start();
    }

    /**
     * garbage collector for the suprised rectangle (animation)
     */
    private void surprisedRectanglesClean() {
        int i = 0;
        while(i < surprisedRectangles.size())
        {
            if(surprisedRectangles.get(i).isAnimationFinished())
                surprisedRectangles.remove(i);
            else
                i++;
        }
    }



/**
 *
 * @return a LaTeX code corresponding to all the proof in this panel
 */
    public String getLaTEXCode()
    {
        String s = "";

        for(ProofFormulaNode node : nodes)
        {
            if(node.noFather())
            {
                s = s + "$$" + node.getLaTEXCodeArbre() + "$$" + "\n" + "\n";
            }
        }

        return s;
    }



    /**
     * select exactly all the node that are located in the rectangle selectionRectangle
     * @param selectionRectangle
     */
    private void selectionnerRectangle(Rectangle selectionRectangle) {
        nodesSelected.clear();

        for(ProofFormulaNode node : getShownNodes())
        {
            if(node.getRectangleFormula().intersects(selectionRectangle))
                nodesSelected.add(node);
        }

        selectionChange();
    }



    /**
     *
     * @param node
     * @return true iff the node is selected
     */
    private boolean isNodeSelected(ProofFormulaNode node) {
        return nodesSelected.contains(node);
    }


    /**
     * compute the positions of the nodes
     */
    private void positionsCompute() {
        for(ProofFormulaNode node : nodes)
        {
            if(node.noFather())
            {
                node.positionsCompute();
            }
        }
    }



    /**
     * s'occupe de vérifier que le timer timerAnimation est activé quand il y a une animation
     * à afficher
     * sinon, éteint le timer timerAnimation (pour pas bouffer tout le temps processeur
     * si il n'y a pas d'animation à faire
     */

    private void animationTimerVerifier() {
        for(ProofFormulaNode node : nodes)
        {
            if(node.isAnimated())
            {
                if(!timerAnimation.isRunning())
                      timerAnimation.start();
                return;
            }
        }

         if(surprisedRectangles.isEmpty())
            timerAnimation.stop();
         else
             if(!timerAnimation.isRunning())
                 timerAnimation.start();
    }




    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem mnuCacherAfficher;
    private javax.swing.JPopupMenu popupMenuNode;
    // End of variables declaration//GEN-END:variables

    private Iterable<ProofFormulaNode> getShownNodes() {
        Set<ProofFormulaNode> shownNode = new HashSet<ProofFormulaNode>();

        for(ProofFormulaNode n : nodes)
        {
            if(n.noFather())
                shownNode.add(n);
            else
            {
                if(!n.getFather().isTreeHidden())
                    shownNode.add(n);
            }
            
        }

        return shownNode;
    }



    /**
     *
     * @param hypothesisFormulas
     * @param goalFormula
     * @return un noeud dont la formule en bas est goalformula et donc les hypothèses
     * principales sont incluses dans l'ensemble hypothesisFormulas
     * S'il n'existe pas un tel noeud, retourne null.
     */
    public ProofFormulaNode getProofNode(Set<Formula> hypothesisFormulas, Formula goalFormula) {
        for(ProofFormulaNode node : nodes)
        {
            ProofFormulaNodeNatDet n = (ProofFormulaNodeNatDet) node;
            if(n.noFather() & n.isTreeOK())
            {
                if(n.getFormula().equals(goalFormula) & hypothesisFormulas.containsAll(n.getHypothesesPrincipalesArbre()))
                    return n;
            }
        }

        return null;



        
    }



    /**
     *
     * @param node
     * @return vrai ssi le noeud node n'a pas d'ancêtre sélectionné
     */
    private boolean noAncestorSelected(ProofFormulaNode node) {
        ProofFormulaNode father = node.getFather();

        while(father != null)
        {
            if(nodesSelected.contains(father))
                return false;

             father = father.getFather();
        }

        return true;
    }


    /**
     * write the proofs of this panel in a file whose name is fileName.
     * The format of saving is an open-source XML format but not a standard format.
     *
     * @param fileName
     * @throws IOException
     */
    void saveInFile(String fileName) throws IOException {
        XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());
        Document document = getDocument();

        sortie.output(document, new FileOutputStream(fileName));


    }





    /**
     *
     * @param nomFichier
     * @return read the file nomFichier and produce a JDOM Document object (which represents the XML tree of the file)
     * @throws JDOMException
     * @throws IOException
     */
    private static Document getDocumentDuFichier(String nomFichier) throws JDOMException, IOException
    {
        SAXBuilder sxb = new SAXBuilder(false);

       //ça c'est un objet qui résout les liens vers les fichiers écrits dans
        //un fichier XML. (à ce que j'ai compris)
        //avec cet objet on peut rediriger la lecture de fichier
        //vers d'autres fichiers.
        //ici on va rediriger la lecture du fichier DTD d'internet
        //vers un InputStream vide !
       EntityResolver entityResolver = new EntityResolver() {

            public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException {

                //quand on tombe sur la DTD référée avec du http etc.
                // on dit en fait qu'on lit dans un InputStream vide !
                if (systemId.startsWith("http://")) {
                         return new InputSource(new InputStream() {
                        @Override
                        public int read() throws IOException {
                            return -1;
                        }
                    });
        }
        return null;

            }
        };

        sxb.setEntityResolver(entityResolver);

        // malheureusement, setValidation(false) fait que la DTD n'est pas utilisée...
        // mais elle est malheureusement quand même parsée. (on vit dans un monde bizarre)

        sxb.setValidation(false);
        return sxb.build(new File(nomFichier));

    }



/**
 * clear the proof panel and read the file fileName written in the open-source
 * non-standard XML format.
 * @param fileName
 * @throws JDOMException
 * @throws IOException
 */
    public void openFromFile(String fileName) throws JDOMException, IOException
    {
        clear();
        openFromDocument(getDocumentDuFichier(fileName));
        
    }



/**
 *
 * @return the JDOM object representing the nodes of the panel.
 * the JDOM object is a representation in XML (open-source but non-standard) format
 */
    private Document getDocument() {
        Element racine = new Element("proofs");
        Document document = new Document(racine);

        for(ProofFormulaNode node : nodes)
        if(node.noFather())
        {
              racine.addContent(node.saveRecursive());
        }

        return document;
    }



    /**
     * add the nodes represented in the XML JDOM object document
     * @param document
     */
    private void openFromDocument(Document document) {
        Element racine = document.getRootElement();

        for(Object elementPartieO : racine.getChildren("proofnode"))
        {
            traiterProofNodeFromXML((Element) elementPartieO);
        }



    }


    /**
     * 
     * @param element
     * @return the node represented by the XML JDOM object element.
     * This function adds ALL the children written in the element and also the
     * root. More precisely, this function returns the root.
     * (it is used as a recursive function)
     * 
     */
    private ProofFormulaNode traiterProofNodeFromXML(Element element) {
        Formula formula = new Formula(element.getChild("formula").getText());

        int x = Integer.parseInt(element.getChild("graphicalposition").getChild("x").getText());
        int y = Integer.parseInt(element.getChild("graphicalposition").getChild("y").getText());
        Point pointMilieu = new Point(x, y);

        ProofFormulaNode node = new ProofFormulaNodeNatDet(pointMilieu, formula);
        for(Object elementChild : element.getChild("children").getChildren())
        {
            ProofFormulaNode nodeChild = traiterProofNodeFromXML((Element) elementChild);
            node.addChild(nodeChild);
        }

        nodeAdd(node);

        return node;
    }



}
